/*
 * Indie Game Jame 2 - moth drawing tool
 * 2004/3/20 Robin Hunicke + many fine jammers
 * 
 */
#include "shared/core/first.h"

// windows crap
#define WIN32_LEAN_AND_MEAN
#define WIN32_EXTRA_LEAN
#define _WIN32_WINNT 0x0500
#include <windows.h>
#include <windowsx.h>

// system
#include <assert.h>
#define _USE_MATH_DEFINES
#include <math.h>

// gl stuff
#include <gl/gl.h>
#include "glutils.h"

// physics stuff
#include "shared/physics/2d/2d_body_hull.h"
#include "shared/physics/2d/2d_body_list.h"
#include "shared/physics/2d/2d_body_box.h"
#include "shared/physics/2d/2d_body_sphere.h"
#include "shared/physics/2d/rigid/2d_rigid_sim.h"
#include "shared/physics/2d/rigid/2d_rigid_drag_force.h"
#include "shared/physics/2d/rigid/2d_rigid_control_force.h"
#include "shared/physics/2d/rigid/2d_rigid_gravity_force.h"
#include "shared/physics/2d/rigid/2d_rigid_spring_force.h"
#include "shared/input/joystick.h"
using namespace ajb;
using namespace Physics::D2;
using namespace Physics::D2::Rigid;

// misc
#include "utils.h"
#include <vector>
#include "shared/input/mouse.h"
using namespace std;

// media lib
#include "music.h"
#include "zsound.h"
#include "bitmap.h"

#include "property_system.h"
#include "property_system_editor.h"
#include "editor.h"

// needed for game consistency on different machines
#include "../../utils/misc/framerate.h"

/********************************* globals **********************************/

// window/camera stuff
char const AppName[] = "Imelda Fiddling";
int WindowWidth = 800;
int WindowHeight = 600;
bool AppQuit = false;  // stuff this to kill yourself
float FOVY = 45.0f, ZNear = 0.1f, ZFar = 1000.0f;  // whatever, we're 2D
float AspectRatio = float(WindowWidth)/float(WindowHeight); // updated in Resize
float CameraAngleRadians = 0.0, CameraX = 0.0, CameraY = 0.0, CameraZ = 100.0;
Vec2 WorldFromWindowCoords( Vec2 window );
Vec2 WindowFromWorldCoords( Vec2 world );

//declarations for drawing boomearng
const int MAX_ARRAY=1000;
const int NUM_COLORS=256;
Vec2 array[NUM_COLORS][MAX_ARRAY];
int drawRead = 0;
int drawWrite = 0;
int drawColor = 0;

bool draw=0;

// win32 stuff
HWND WindowHandle;
HDC WindowDC;

// gl stuff
HGLRC WindowRC;

// functions for you to play in
void Resize( int w, int h );
void Initialize( void );
void Uninitialize( void );
void Idle( void );
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam);

// system
Music music;

// physics stuff
Sim simulator;
Drag_Force    global_drag(.50f, .50f);    //drag of .50, .50 good for boomerang stability
Gravity_Force gravity(Vec2(0.0f, -1.0f));
Rigid_Body    world_body(make_Body_Null(),Vec2(0,0),0,0,0);
Constant_Force  * boomerang_force; //boomerang force - creates elliptical movement


char *TestLevelFileName = "bodies.physdb";

PROP_PARSE_THIS_FILE;
PROP_STRUCT struct global_properties
{
    float Gravity;
    
    float MouseSpringKs;
    float MouseSpringKd;
    float MouseSpringRestLength;
};
static global_properties Globals;
static property *GlobalProperties;
static property_system_editor *GlobalsEditor;

static bool MovingCamera = false;


// game stuff

int strMatchWild( char *srcWild, char *other )
{
	assert( srcWild );
	if( !other )
		return 0;
	char *a = other;
	char *b = srcWild;
	char *bLastStar = 0;
	char *aLastStopper = 0;

	while( *b )
	{
		if( *b == '*' )
		{
			bLastStar = b;
			++b;
			char stopper = tolower(*b);
			while( *a && tolower(*a) != stopper )
			{
				++a;
			}
			if( tolower(*a) != stopper )
			{
				return 0;
			}
			aLastStopper = a;
		}
		else
		{
			if( tolower(*b) != tolower(*a) )
			{
				if( bLastStar )
				{
					b = bLastStar;
					a = aLastStopper + 1;
					continue;
				}
				return 0;
			}
			++a;
			++b;
		}
	}
	return (*a) == '\0';
}
struct game_object {
    
    game_object( char *texture_file , Rigid_Body *rbInit ) :
            rb(rbInit)
        {
            assert(rb);

            simulator.add(*rb);
        }

    ~game_object( void )
        {
            delete rb;
        }

//Ken's code for finding named objects
	int isName( char *nameWild )
	{
		assert( rb );
		return strMatchWild( nameWild, (char*)rb->name().c_str() );
	}

    
    Rigid_Body *rb;
};


static game_object * boomerang; //declare boomerang

static vector<game_object*> objs;
static vector<Binary_Force *> forces;

static Vec2 window_mouse, world_mouse;
static Joystick PS2Controller;

static Binary_Force *spring = 0;


game_object *thingGetByName( char *nameWild )
{
	for( int i=0 ; i<objs.size() ; ++i )
	{
		game_object *thing = objs[i];
		assert( thing );
		if( thing->isName( nameWild ) )
		{
			return thing;
		}
	}
	return 0;
}
GLuint
LoadTexture(char *Filename)
{
    GLuint texture;
    
    Bitmap_Load_Result result;
    verify(LoadBitmapByExtension(Filename,&result));

    // gl stuff
    texture = gl_texture_from_bitmap(result.data, result.width, result.height,
                                     result.bytes_per_pixel);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);

    // delete the bitmap
    delete [] result.data;

    return(texture);
}

void
glBody1v(Body *body, bool Solid)
{
    if(body)
    {
		{
			Body_Hull *p = dynamic_cast<Body_Hull*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Box *p = dynamic_cast<Body_Box*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				for(int i = 0;i<p->num_verts();++i)
				{
					Vec2 worldspace = p->vert(i);
					glVertex2f(worldspace.x(),worldspace.y());
				}
				glEnd();
			}
		}

		{
			Body_Sphere *p = dynamic_cast<Body_Sphere*>(body);
			if(p)
			{
				glBegin(Solid ? GL_TRIANGLE_FAN : GL_LINE_LOOP);
				int total_steps = 16;
				Vec2 center = p->x();
				if(Solid)
				{
					glVertex2f(center.x(), center.y());
				}
				for(int a = 0;a<total_steps;++a)
				{
					float Angle = 2.0f * 3.14159265358979323846f * ((float)a / (float)(total_steps - 1));
					float Radius = p->r();
					Vec2 arm(Radius*cos(Angle), Radius*sin(Angle));
					arm += center;
					glVertex2f(arm.x(), arm.y());
				}
				glEnd();
			}
		}

        Body_List *l = dynamic_cast<Body_List*>(body);
        if(l)
        {
            {for(int Index = 0;
                 Index < l->num_bodies();
                 ++Index)
            {
                glBody1v(&l->body(Index), Solid);
            }}
        }
    }
}

void
glForce1v(Binary_Force *Force)
{
    if(Force)
    {
        glBegin(GL_LINES);
        glVertex2f(Force->m_a.x_worldS().x(), Force->m_a.x_worldS().y());
        glVertex2f(Force->m_b.x_worldS().x(), Force->m_b.x_worldS().y());
        glEnd();
    }
}

void
ClearMouseSpring(void)
{
    if(spring)
    {
        simulator.remove(*spring);
        delete spring;
        spring = 0;
    }
}

void
ClearAll(void)
{
    ClearMouseSpring();
    
    for(int i = 0;i < objs.size();++i)
    {
        assert(objs[i]);
        simulator.remove(*objs[i]->rb);
    }

    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        simulator.remove(*forces[i]);
    }

    forces.clear();
    objs.clear();
}

void
LoadTestFile(void)
{
    int num_rb=0, num_force=0;
    Rigid_Body **rb_list;
    Binary_Force **force_list;

    loadPhysicsFile(TestLevelFileName, -3,-5, 1.0/8, 1.0/16, &rb_list, &num_rb, &force_list, &num_force, &world_body);

    for (int i=0; i < num_rb; ++i)
    {
        objs.push_back(new game_object("content/bitmaps/test.jpg",
                                       rb_list[i]));
    }

    for (int i=0; i < num_force; ++i)
    {
        forces.push_back(force_list[i]);
        simulator.add(*force_list[i]);
    }

	//ellipse code
	//find boomerang 
    game_object *obj=thingGetByName("boomerang");
	//name the boomerang object
	boomerang=obj;

	if (obj !=NULL) 
	{
		obj->rb->v=Vec2(0,0); //halt boomerang from previous load
		obj->rb->v=Vec2(0,50); //give it a push

		boomerang_force=new Constant_Force (Handle (*obj->rb,Vec2(0,0))); //create some force
		simulator.add(*boomerang_force);//add force to the world
	}
	

}

/******************************** Initialize ********************************/
/* the window and graphics subsystem are set up but still hidden            */
void Initialize( void )
{
    InitializeFrameRate(60.0f);

	PS2Controller.init();

    Globals.Gravity = 0.0f;

    Globals.MouseSpringKs = 128.0f;
    Globals.MouseSpringKd = 32.0f;
    Globals.MouseSpringRestLength = 0.0f;

    GlobalProperties = CloneProperty(PROP_LOCAL_STRUCT(global_properties));
    ConvertStructToProperty(PROP_LOCAL_STRUCT(global_properties), &Globals,
                            GlobalProperties);
    GlobalsEditor = OpenPropertySystemEditor("Globals", 850, 50, 400, 400);
    EditPropertiesInPlace(GlobalsEditor, GlobalProperties);
    
    GetTime(); // throw away initial 0.0 so we're seconds-from-startup

    // gl stuff
    glClearColor(1.0f,1.0f,1.0f,0.0f);
    glEnable(GL_CULL_FACE);
    glEnable(GL_NORMALIZE);
    glEnable(GL_DEPTH_TEST);
    glShadeModel(GL_SMOOTH);

    // init our view
    Resize(WindowWidth,WindowHeight);

    // music init
    music.init();
    music.set_subdirectory("content/music/");
    music.setupnotes();
    music.loadseg("background", "simple_loop.sgt", MUSIC_TRACK_MAIN);

    // sound init
    zsoundInit(WindowHandle);

    // initialize the physics
    simulator.add(global_drag);
    simulator.add(gravity);
    simulator.add(world_body);

    LoadTestFile();
}

/*********************************** Idle ***********************************/
/* called by the message loop, and by paint messages                        */
void Idle( void )
{
    ConvertPropertyToStruct(FindSubPropertyByName(LocalTemplates, "global_properties"),
                            GlobalProperties,
                            &Globals);

    gravity.m_g.set_y(Globals.Gravity);
    
	PS2Controller.sample();
	Vec2 LeftStick = PS2Controller.left_stick();
	Vec2 RightStick = PS2Controller.right_stick();

	/**************************** step the simulator ****************************/   
	
	if (spring)                             //update
        spring->m_a = Handle(world_body, world_mouse);

	//setting boomerang_force
	Vec2 center(5,0);
	Vec2 delta=(center-boomerang->rb->x);
	boomerang_force->m_f=(2.0 * delta);
	
    int const NumberOfSimSteps = 1;
    for(int StepCounter = 0;StepCounter < NumberOfSimSteps;++StepCounter) {
        // step as many times as you like
        simulator.step(1/(NumberOfSimSteps*GetFrameRate()));
        // do any game stuff you want here (like per-step collision response, etc.)
    }
    // now wait for the rest of the time so we're consistent on fast machines
    WaitFrameRate(GetFrameRate());

    /**************************** render everything *****************************/
   
    // lots of gl stuff
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDisable(GL_DEPTH_TEST);

    glLoadIdentity();
    glRotated(180.0*(-CameraAngleRadians/M_PI),0,0,1);
    glTranslatef(-CameraX,-CameraY,-CameraZ);

    // draw axes
    glBegin(GL_LINES);
    glColor3f(1,0,0);  glVertex2f(0,0);  glVertex2f(1,0);
    glColor3f(0,1,0);  glVertex2f(0,0);  glVertex2f(0,1);
    glEnd();

    // draw mouse
    glPointSize(4.0);
    glBegin(GL_POINTS);
    glColor3f(0,0,0);  glVertex2fv(world_mouse.m_v);
    glEnd();

	//boomerang draw code
	//stuff location of boomerang into array
	if (draw)
	{
		array[drawColor][drawWrite] = boomerang->rb->x;
		drawWrite = (drawWrite+1) % MAX_ARRAY;
		drawRead++;
	}

	//random color lookup
	static unsigned char colors[256][3] = {0,};
	if( colors[0][0]== 0 ) {
		memset( array, 0, sizeof(array) );
		for( int i=0; i<256; i++ ) {
			colors[i][0] = rand() % 256;
			colors[i][1] = rand() % 256;
			colors[i][2] = rand() % 256;
		}
	}

    //draw the path of the boomerang
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	for (int c=0; c<=drawColor; c++)
	{
		//glColor3ub(colors[c][0],colors[c][1],colors[c][2]);
		glLineWidth(4.f);
		glBegin(GL_TRIANGLE_FAN);
		for (int i=0; i<MAX_ARRAY; i++)
		{
			glColor4f((float)colors[c][0]/255.f,(float)colors[c][1]/255.f,(float)colors[c][2]/255.f,(float)i/(float)MAX_ARRAY);
			//glColor3ub(colors[c][0],colors[c][1],colors[c][2]);
			glVertex2fv(array[c][(i+drawRead)%MAX_ARRAY].m_v);
		}
		glEnd();
	}


	//reset line width
	glLineWidth(1.f);

	
	// iterate all the bodies, rendering them
    glColor3f(0,1,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj.rb->m_Pbody.get(), true);
    }

    glColor3f(0,0,0);
    for(int i = 0;i < objs.size();++i) 
	{
        assert(objs[i]);
        game_object &obj = *objs[i];

        glBody1v(obj.rb->m_Pbody.get(), false);
    }
    
    // iterate all the forces, rendering them
    glColor3f(1,0,0);
    for(int i = 0;i < forces.size();++i)
    {
        assert(forces[i]);
        Binary_Force *Force = forces[i];

        glForce1v(Force);
    }
    glForce1v(spring); 

    glDisable(GL_BLEND);

	glColor3f(0, 0, 1);
	glBegin(GL_LINES);
	float StickScale = 10.0f;
	glVertex2f(0, 0);
	glVertex2f(StickScale * LeftStick.x(), StickScale * LeftStick.y());
	glVertex2f(0, 0);
	glVertex2f(StickScale * RightStick.x(), StickScale * RightStick.y());
	glEnd();

    glColor3f(0, 0, 0);
    char buf[100];
    sprintf(buf,"wm: %5.1f, %5.1f    sm: %5.1f, %5.1f",
            world_mouse.x(),world_mouse.y(),window_mouse.x(),window_mouse.y());
    DrawString(50,50,buf);

	char *ButtonStrings[] =
	{
		"BUTTON_RIGHT_D",
		"BUTTON_RIGHT_U",
		"BUTTON_RIGHT_L",
		"BUTTON_RIGHT_R",
		"BUTTON_SHOULDER_R1",
		"BUTTON_SHOULDER_R2",
		"BUTTON_SHOULDER_L1",
		"BUTTON_SHOULDER_L2",
		"BUTTON_RSTICK",
		"BUTTON_LSTICK",
		"BUTTON_START",
		"BUTTON_SELECT",
	};

	int Y = 70;
	for(int ButtonIndex = 0;
		ButtonIndex < Joystick::NUM_BUTTONS;
		++ButtonIndex)
	{
		if(PS2Controller.button_is_down(ButtonIndex))
		{
			DrawString(50,float(Y),ButtonStrings[ButtonIndex]);
			Y += 10;
		}
	}



    DrawTimingBars();
    SwapBuffers(WindowDC);
}

/******************************** AppWndProc ********************************/
/* lame to expose this, but the alternatives all suck too                   */
LONG FAR PASCAL AppWndProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{
    switch (msg)
    {
        // keyboard
        case WM_CHAR:
            break;

        case WM_MOUSEWHEEL:
        {
            float d = float((short)HIWORD(wParam))/float(WHEEL_DELTA);
            CameraZ *= 1.0f + d/10.0f;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }
        case WM_LBUTTONDOWN:
        {
			SetCapture(hwnd);

            Vec2 body_space_position;
            Rigid_Body* rb = simulator.pick(body_space_position, world_mouse);
            if (rb)
            {
                spring = new Spring_Force(
                    Handle(world_body, world_mouse),
                    Handle(*rb, body_space_position), 
                    Globals.MouseSpringKs,
                    Globals.MouseSpringKd,
                    Globals.MouseSpringRestLength
                    );
                simulator.add(*spring);
            }

        } break;

        case WM_LBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_RBUTTON))
			{
				ReleaseCapture();
			}

            ClearMouseSpring();
        } break;

        case WM_RBUTTONDOWN:
        {
			SetCapture(hwnd);

			if(spring)
            {
                forces.push_back(spring);
                spring = 0;
            }
            break;
        }
	
        case WM_MBUTTONDOWN:
        {
            MovingCamera = true;
        } break;
        case WM_MBUTTONUP:
        {
            MovingCamera = false;
        } break;

        case WM_RBUTTONUP:
        {
			if(!(wParam & MK_MBUTTON & MK_LBUTTON))
			{
				ReleaseCapture();
			}
        } break;
            
        case WM_KEYDOWN: {
            switch(wParam) {
                case VK_LEFT: {
                    CameraAngleRadians += float((5.0/180.0)*M_PI);
                    break;
                }
                case VK_RIGHT: {
                    CameraAngleRadians -= float((5.0/180.0)*M_PI);
                    break;
                }

                case 'M': {
                    // demo the audio stuff
                    if(0x8000 & GetKeyState(VK_SHIFT)) {
                        music.play("background");
                    } else {
                        zsoundPlay("content/sounds/bad.0002.wav");
                    }
                    break;
                }
				
				// zip to origin if spazzing
				case 'Z': {
					boomerang->rb->v.set_x(0);
					boomerang->rb->v.set_y(0);

					boomerang->rb->x.set_x(0);
					boomerang->rb->x.set_y(0);

					boomerang->rb->w=1;
					break;
                } 

				//toggle drawing on and off
				case 'D': {
					draw=!draw;
					break;
				}

				//switch colors
				case 'F': {
					drawRead = 0;
					drawColor = (drawColor+1) % NUM_COLORS;
					break;
				}

						  

                case 'E': {
                    launchEditor(TestLevelFileName);
                    ClearAll();
                    LoadTestFile();
                } break;
            
                case VK_ESCAPE: {
                    AppQuit = true;
                    break;
                }
            }
            break;
        }
        case WM_SYSKEYDOWN:
        case WM_SYSKEYUP:
        case WM_KEYUP:
            break;

        case WM_MOUSEMOVE: {
			Vec2 new_window_mouse = Vec2(float(GET_X_LPARAM(lParam)),float(GET_Y_LPARAM(lParam)));
			if(MovingCamera)
			{
                Vec2 d_world = WorldFromWindowCoords(new_window_mouse) - world_mouse;
				CameraX -= d_world.x();
				CameraY -= d_world.y();
			}
            window_mouse = new_window_mouse;
            world_mouse = WorldFromWindowCoords(window_mouse);
            break;
        }

         
            /**************** below is stuff you probably don't care about ****************/
         
        case WM_DESTROY: {
            PostQuitMessage(0);
            break;
        }
        case WM_PAINT: {
            PAINTSTRUCT ps;
            BeginPaint(hwnd,&ps);
            Idle();
            EndPaint(hwnd,&ps);
            return 0L;
        }
        case WM_SIZE: {
            Resize(LOWORD(lParam),HIWORD(lParam));
            break;
        }
    }
    return (LONG)DefWindowProc(hwnd,msg,(LONG)wParam,(LONG)lParam);
}

/********************************** Resize **********************************/
/* called on WM_SIZE, and at the bottom of initialize                       */
void Resize( int w, int h )
{
    // compute our frustrum planes, the x,y plane values are on the near clip plane
    WindowWidth = w;
    WindowHeight = h;
    AspectRatio = float(w)/float(h);
    float ymax = ZNear * float(tan(FOVY/2.0));
    float ymin = -ymax;
    float xmin = AspectRatio * ymin;
    float xmax = AspectRatio * ymax;
   
    // load the viewport and projection matrix, gl stuff
    glViewport(0,0,w,h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    // gluPerspective doesn't set the near plane as the projection
    // plane, so you can't do good math, so do it ourselves
    glFrustum(xmin,xmax,ymin,ymax,ZNear,ZFar);
    glMatrixMode(GL_MODELVIEW);
}

/******************************* Uninitialize *******************************/
/* called before the graphics is destroyed, but the window is already toast */
void Uninitialize( void )
{
}

/********************************* WinMain **********************************/

int PASCAL WinMain(HINSTANCE hInst, HINSTANCE hPrev, LPSTR szCmdLine, int sw)
{
    // register and create our window
    WNDCLASS cls;
    if(!hPrev) {
        cls.hCursor        = LoadCursor(NULL,IDC_ARROW);
        cls.hIcon          = 0;
        cls.lpszMenuName   = 0;
        cls.lpszClassName  = AppName;
        cls.hbrBackground  = 0;
        cls.hInstance      = hInst;
        cls.style          = CS_VREDRAW | CS_HREDRAW | CS_DBLCLKS | CS_OWNDC;
        cls.lpfnWndProc    = (WNDPROC)AppWndProc;
        cls.cbWndExtra     = 0;
        cls.cbClsExtra     = 0;

        if (!RegisterClass(&cls))
            return FALSE;
    }

    DWORD Style = WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN | WS_CLIPSIBLINGS;
    RECT WindowRect = { 0, 0 };
    WindowRect.right = WindowWidth;
    WindowRect.bottom = WindowHeight;
    BOOL Menu = FALSE;

    AdjustWindowRect(&WindowRect,Style,Menu);
    
    WindowHandle = CreateWindow(AppName,AppName,Style,
                                0,50,
                                WindowRect.right-WindowRect.left,
                                WindowRect.bottom-WindowRect.top,
                                0,0,hInst,0);

    // create our gl stuff
    WindowDC = GetDC(WindowHandle);
    PIXELFORMATDESCRIPTOR FormatDescriptor = {
        sizeof (PIXELFORMATDESCRIPTOR), 1,
        PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER,
        PFD_TYPE_RGBA, 32,   // 32bpp
        0,0, 0,0, 0,0, 0,0,
        0, 0,0,0,0,
        24, 8,               // depth, stencil
        0,
        PFD_MAIN_PLANE,
        0,
        0,0,0
    };
    int FormatIndex = ChoosePixelFormat(WindowDC,&FormatDescriptor);
    SetPixelFormat(WindowDC,FormatIndex,&FormatDescriptor);
    WindowRC = wglCreateContext(WindowDC);
    wglMakeCurrent(WindowDC,WindowRC);

    // okay, ready to go
   
    Initialize();    // call the app
   
    ShowWindow(WindowHandle,sw);
   
    // pump the message queue
    MSG msg;
    while(!AppQuit) {
        if(PeekMessage(&msg,0,0,0,PM_REMOVE)) {
            if(msg.message == WM_QUIT)
                break;
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        } else {
            Idle();    // call the app
        }
    }

    Uninitialize();  // call the app
   
    // clean up gl stuff and exit
    wglMakeCurrent(0,0);
    wglDeleteContext(WindowRC);
    ReleaseDC(WindowHandle,WindowDC);
   
    return int(msg.wParam);
}

/********************** utilities that use our globals **********************/
Vec2 WorldFromWindowCoords( Vec2 window )
{
    float xn = float(window.x())/WindowWidth - 0.5f;
    float yn = float(WindowHeight-window.y())/WindowHeight - 0.5f;
    float xc = float(xn*2.0*CameraZ*tan(FOVY/2)*AspectRatio);
    float yc = float(yn*2.0*CameraZ*tan(FOVY/2));
    return Vec2(CameraX,CameraY) + Matrix22::rotation(CameraAngleRadians) * Vec2(xc,yc);
}
Vec2 WindowFromWorldCoords( Vec2 world )
{
    Vec2 c = Matrix22::rotation(-CameraAngleRadians) * (world - Vec2(CameraX,CameraY));
    float xn = float(c.x()/(2.0*CameraZ*tan(FOVY/2)*AspectRatio));
    float yn = float(c.y()/(2.0*CameraZ*tan(FOVY/2)));
    float xs = WindowWidth*(xn + 0.5f);
    float ys = WindowHeight - WindowHeight*(yn + 0.5f);
    return Vec2(xs,ys);
}
